@@ -1,7 +1,7 @@
 /*
  * Output command-line help to stdout.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -9,6 +9,7 @@
 #include "config.h"
 #include "pv.h"
 #include <stdio.h>
+#include <signal.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
@@ -306,6 +307,9 @@ void display_help(void)
 		{ "-8", "--bits", NULL,
 		 N_("show number of bits transferred"),
 		 { 0, 0, 0, 0} },
+		{ "-k", "--si", NULL,
+		 N_("treat suffixes as multiples of 1000 rather than 1024"),
+		 { 0, 0, 0, 0} },
 		{ "-T", "--buffer-percent", NULL,
 		 N_("show percentage of transfer buffer in use"),
 		 { 0, 0, 0, 0} },
@@ -383,11 +387,11 @@ void display_help(void)
 		{ "-X", "--discard", NULL,
 		 N_("discard input instead of writing to output"),
 		 { 0, 0, 0, 0} },
-#ifdef HAVE_IPC
+#ifdef SA_SIGINFO
 		{ "-R", "--remote", N_("PID"),
 		 N_("update settings of process PID"),
 		 { 0, 0, 0, 0} },
-#endif				/* HAVE_IPC */
+#endif				/* SA_SIGINFO */
 		{ "", NULL, NULL, NULL, { 0, 0, 0, 0} },
 		{ "-P", "--pidfile", N_("FILE"),
 		 N_("save process ID in FILE"),
